static char  VERSION_STRING[]	= "@(#)pcal v4.1 - generate Postscript calendars";
/*
 * pcal.c - generate PostScript file to print calendar for any month and year
 *
 * The original PostScript code to generate the calendars was written by
 * Patrick Wood (Copyright (c) 1987 by Patrick Wood of Pipeline Associates,
 * Inc.), and authorized for modification and redistribution.  The calendar
 * file inclusion code was originally written in "bs(1)" by Bill Vogel of
 * AT&T.  Patrick's original PostScript was modified and enhanced several
 * times by King Ables, Tom Tessin, and others whose names have regrettably 
 * been lost.  This C version was originally created by Ken Keirnan of Pacific
 * Bell; additional enhancements by Joseph P. Larson, Ed Hand, Andrew Rogers, 
 * Mark Kantrowitz, and Joe Brownlee.  The moon routines were originally 
 * written by Jef Poskanzer and Craig Leres, and were incorporated into Pcal
 * by Richard Dyson.
 *
 * Contents:
 *
 *		alt_fopen
 *		color_msg
 *		get_args
 *		get_flag
 *		main
 *		set_color
 *		usage
 *
 * Revision history:
 *
 *	4.1	AWR	08/02/91	Add -G flag to print "gray" dates as
 *					outlined, gray-filled characters
 *
 *					Fix potential bug in julday() (cf.
 *					moonphas.c)
 *
 *	4.02	AWR	07/02/91	Add -v flag to print version info only;
 *					call find_executable() to get true
 *					program pathname (cf. pcalutil.c);
 *					add format specifiers to text strings
 *					(cf. writefil.c)
 *
 *	4.01	AWR	03/19/91	Incorporate revised moonphas.c (q.v.)
 *
 *	4.0	AWR	02/24/91	Add alt_fopen() to search for file
 *					in alternate path; use to look for
 *					date file in same directory as
 *					Pcal executable (as per Floyd Miller)
 *
 *					Support negative ordinals (cf.
 *					readfile.c, pcalutil.c)
 *
 *					Support expressions in preprocessor
 *					"if{n}def" lines (cf. exprpars.c)
 *
 *					Support "even", "odd" ordinals (cf.
 *					readfile.c) and ordinals > 5th
 *
 *					Support -B (leave unused boxes blank)
 *					flag
 *
 *					Separated into moonphas.c, pcal.c,
 *					pcalutil.c, readfile.c, and writefil.c;
 *					added support for moon phase file
 *
 *					Support -w (whole year) flag; fix
 *					various bugs and nonportable constructs
 *
 *	3.0	AWR	12/10/90	Support concept of "weekday", "workday",
 *					and "holiday" (and converses)
 *
 *					Substantial revision of program logic:
 *					extracted pcaldefs.h and pcallang.h,
 *					moving all language dependencies (even
 *					flag names) to the latter.
 *
 *					add -I flag to reinitialize all
 *	 				flags to program defaults; -j and -J
 *					flags (print Julian dates); add -x,
 *				 	-y, -X, -Y flags (as per Ed Hand)
 *					for output scaling and translation
 *
 *					allow "wildcard" dates (e.g., "all
 *					Thursday{s} in Oct", "last Friday in
 *					all") and notes ("note all <text>);
 *					print full "help" message (including
 *					date file syntax)
 *
 *	2.6	AWR	10/15/90	parse floating dates (e.g. "first
 *					Monday in September") and relative
 *					floating dates (e.g., "Friday after
 *					fourth Thursday in November"); simplify
 *					logic of -F option; add -F to usage 
 *					message; replace COLOR_MSG() string
 *					with color_msg() routine; add -h
 *					(help message) and -A | -E (American |
 *					European date format) flags; renamed
 *					flag sets for clarity; more comments
 *
 *	2.5	JAB	10/04/90	added -F option
 *
 *	2.4	---	10/01/90	* no modifications *
 *
 *	2.3	JWZ/AWR	09/18/90	added moon routines
 *
 *	2.2	AWR	09/17/90	revise logic of parse(); new usage
 *					message
 *
 *		JAB/AWR	09/14/90	support "note" lines in date file
 *
 *	2.1	MK/AWR	08/27/90	support -L, -C, -R, -n options;
 *					print holiday text next to date
 *
 *		AWR	08/24/90	incorporate cpp-like functionality;
 *					add -D and -U options; save date file
 *					information in internal data structure;
 *					look for PCAL_OPTS and PCAL_DIR; look
 *					for ~/.calendar and ~/calendar
 *
 *	2.0	AWR	08/08/90	included revision history; replaced -r
 *					flag with -l and -p; replaced -s and -S
 *					flags with -b and -g; recognize flags
 *					set in date file; translate ( and ) in
 *					text to octal escape sequence; usage()
 *					message condensed to fit 24x80 screen
 *
 *	Parameters:
 *
 *		pcal [opts]		generate calendar for current month/year
 *					(current year if -w flag specified)
 *
 *		pcal [opts] yy		generate calendar for entire year yy
 *
 *		pcal [opts] mm yy	generate calendar for month mm
 *					(Jan = 1), year yy (19yy if yy < 100)
 *					(12 months starting with mm/yy if -w
 *					specified)
 *
 *		pcal [opts] mm yy n	as above, for n consecutive months (n
 *					rounded to next multiple of 12 if -w
 *					specified)
 *
 *	Output:
 *
 *		PostScript file to print calendars for all selected months.
 *
 *	Options:
 *
 *		-I		initialize all parameters to program defaults
 *
 *		-b <DAY>	print specified weekday in black
 *		-g <DAY>	print specified weekday in gray
 *				(default: print Saturdays and Sundays in gray)
 *
 *		-G		print "gray" dates as filled outlines
 *		-O		print "gray" dates as unfilled outlines
 *				(default: gray characters)
 *
 *		-d <FONT>	specify alternate font for day names
 *				(default: Times-Bold)
 *
 *		-n <FONT>	specify alternate font for notes in boxes
 *				(default: Helvetica-Narrow)
 *
 *		-t <FONT>	specify alternate font for titles
 *				(default: Times-Bold)
 *
 *		-D <SYM>	define preprocessor symbol
 *		-U <SYM>	un-define preprocessor symbol
 *
 *		-e		generate empty calendar (ignore date file)
 *
 *		-f <FILE>	specify alternate date file (default:
 *				~/.calendar on Un*x, SYS$LOGIN:CALENDAR.DAT
 *				on VMS, s:calendar.dat on Amiga; if
 *				environment variable [logical name on VMS]
 *				PCAL_DIR exists, looks there instead; if
 *				not found in either place, looks in same
 *				directory as Pcal executable)
 *
 *		-o <FILE>	specify alternate output file (default:
 *				stdout on Un*x, CALENDAR.PS on VMS, 
 *				RAM:calendar.ps on Amiga)
 *
 *		-L <STRING>	specify left foot string   (default: "")
 *		-C <STRING>	specify center foot string (default: "")
 *		-R <STRING>	specify right foot string  (default: "")
 *
 *		-l		generate landscape-mode calendars
 *		-p		generate portrait-mode calendars
 *				(default: landscape-mode)
 *
 *		-h		(command line only) write version information
 *				and "usage" message to stdout
 *		-v		(command line only) write version information
 *				alone to stdout
 *
 *		-m		draw a small moon icon on the days of the
 *				full, new, and half moons.
 *		-M		draw a small moon icon every day.  
 *				(default: no moons)
 *
 *		-F <DAY>	select alternate day to be displayed as the 
 *				first day of the week (default: Sunday)
 *
 *		-A		dates are in American format (e.g., 10/15/90,
 *				Oct 15) (default)
 *		-E		dates are in European format (e.g., 15.10.90,
 *				15 Oct)
 *
 *		-x <XSCALE>	These two options can be used to change
 *		-y <YSCALE>	the size of the calendar.
 *
 *		-X <XTRANS>	These two options can be used to relocate
 *		-Y <YTRANS>	the position of the calendar on the page.
 *
 *		-j		print Julian dates (day of year)
 *		-J		print Julian dates and days remaining
 *				(default: neither)
 *
 *		-w		print whole year (12 months) per page
 *
 *		-B		leave unused calendar boxes blank
 *
 *
 *	There are many ways to specify these options in addition to using the
 *	command line; this facilitates customization to the user's needs.
 *
 *	If the environment variable (global symbol on VMS) PCAL_OPTS is
 *	present, its value will be parsed as if it were a command line.
 *	Any options specified will override the program defaults.
 *
 *	All but the -e, -f, -h, -v, -D, and -U options may be specified in the
 *	date file by including one or more lines of the form "opt <options>".
 *	Any such options override any previous values set either as program
 *	defaults, via PCAL_OPTS, or in previous "opt" lines.
 *
 *	Options explicitly specified on the command line in turn override all
 *	of the above.
 *
 *	Any flag which normally takes an argument may also be specified without
 *	an argument; this resets the corresponding option to its default.  -D
 *	alone un-defines all symbols; -U alone has no effect.
 *
 *	Parameters and flags may be mixed on the command line.  In some cases
 *	(e.g., when a parameter follows a flag without its optional argument)
 *	this may lead to ambiguity; the dummy flag '-' (or '--') may be used
 *	to separate them, i.e. "pcal -t - 9 90".
 *
 *
 *	Date file syntax:
 *
 *	The following rules describe the syntax of date file entries:
 *
 *	  year <year>
 *
 *	  opt <options>
 *
 *	  note <month_spec> <text>
 *	  note <month> <text>
 *
 *	  if -A flag (American date formats) specified:
 *	    <month_name> <day>{*} {<text>}
 *	    <month><sep><day>{<sep><year>}{*} {<text>}
 *
 *	  if -E flag (European date formats) specified:
 *	    <day> <month_name>{*} {<text>}
 *	    <day><sep><month>{<sep><year>}{*} {<text>}
 *
 *	  <ordinal> <day_name> in <month_spec>{*} {<text>}
 *	  <day_name> <prep> <date_spec>
 *
 *	where
 *
 *	  {x}	  means x is optional
 *
 *	  <date_spec> := any of the above date specs (not year, note, or opt)
 *	  <month_name> := first 3+ characters of name of month, or "all"
 *	  <month_spec> := <month_name>, or "year"
 *	  <day_name> := first 3+ characters of name of weekday, "day",
 *			"weekday", "workday", "holiday", "nonweekday",
 *			"nonworkday", or "nonholiday"
 *	  <ordinal> := ordinal number ("1st", "2nd", etc.), "first" .. "fifth",
 *			"last", "even", "odd", or "all"
 *	  <prep> := "before", "preceding", "after", "following", "on_or_before",
 *			or "on_or_after"
 *	  <sep> := one or more non-numeric, non-space, non-'*' characters
 *	  <month>, <day>, <year> are the numeric forms
 *
 *	  <options> := any command-line option except -e, -f, -h, -v, -D, -U
 *
 *	Comments start with '#' and run through end-of-line.
 *
 *	Holidays may be flagged by specifying '*' as the last character of
 *	the date field(s), e.g. "10/12* Columbus Day", "July 4* Independence
 *	Day", etc.  Any dates flagged as holidays will be printed in gray, and
 *	any associated text will appear adjacent to the date.
 *
 *	Note that the numeric date formats (mm/dd{/yy}, dd.mm{.yy}) support
 *	an optional year, which will become the subsequent default year.  The
 *	alphabetic date formats (month dd, dd month) do not support a year
 *	field; the "year yy" command is provided to reset the default year.
 *
 *	"Floating" days may be specified in the date file as "first Mon in 
 *	Sep", "last Mon in May", "4th Thu in Nov", etc.; any word may be
 *	used in place of "in".  "Relative floating" days (e.g. "Fri after 4th 
 *	Thu in Nov") are also accepted; they may span month/year bounds.
 *	Pcal also accepts date specs such as "all Friday{s} in October", "last
 *	Thursday in all", etc., and produces the expected results; "each" and
 *	"every" are accepted as synonyms for "all".  Negative ordinals are
 *	allowed; "-2nd" means "next to last".
 *
 *	The words "day", "weekday", "workday", and "holiday" may be used as
 *	wildcards: "day" matches any day, "weekday" matches any day normally
 *	printed in black, "workday" matches any day normally printed in black
 *	and not explicitly flagged as a holiday, and "holiday" matches any
 *	day explicitly flagged as a holiday.  "Nonweekday", "nonworkday",
 *	and "nonholiday" are also supported and have the obvious meanings.
 *	
 *	"Odd" and "even" do not refer to the actual date; instead, "odd"
 *	means "alternate, starting with the first"; "even" means "alternate,
 *	starting with the second".  Thus, "odd Fridays in March" refers to
 *	the first, third, and (if present) fifth Fridays in March - not to
 *	those Fridays falling on odd dates.
 *
 *	"All" refers to each individual month; "year" refers to the year
 *	as an entity.  Thus "odd Fridays in all" refers to the first/third/
 *	fifth Friday of each month, while "odd Fridays in year" refers to
 *	the first Friday of January and every other Friday thereafter.
 *
 *	Additional notes may be propagated to an empty calendar box by the
 *	inclusion of one or more lines of the form "note <month> <text>",
 *	where <month> may be numeric or alphabetic; "note all <text>"
 *	propagates <text> to each month in the current year.
 *
 *	Pcal also allows format specifiers in the text (and foot strings -
 *	cf. the -L, -C, and -R options); each will be replaced by its
 *	equivalent string as outlined in the table below.  (Most of these are
 *	derived from the strftime() function; %[lo0+-] are Pcal-specific.)
 *
 *		%a : abbreviated weekday
 *		%A : full weekday
 *		%b : abbreviated month name
 *		%B : full month name
 *		%d : day of month (1-31)
 *		%j : day of year (1-366)
 *		%l : days left in year (0-365)
 *		%m : month (1-12)
 *		%U : week number (0-53)
 *		%W : week number (0-53)
 *		%y : year w/o century (00-99)
 *		%Y : year w/century
 *		%% : '%' character
 *
 *		%o : print number as ordinal
 *		%0 : print number with leading zeroes
 *		%+ : use following month or year
 *		%- : use previous month or year
 *
 *	%U considers the first logical Sunday (the first day of the week as
 *	printed; cf. the -F flag) of the year as the first day of week 1;
 *	%W uses the first logical Monday instead.  This is an extension of
 *	strftime()'s behavior.
 *
 *	%o prints a number as an ordinal, with the appropriate suffix ("st",
 *	"nd", "rd", or "th" in English) appended; for example, "%od" prints
 *	the day of the month as "1st, 2nd, 3rd, ...".
 *
 *	Unlike strftime(), Pcal's default is to print numbers (except %y)
 *	without leading zeroes.  If leading zeroes are desired, the '0'
 *	prefix may be used; for example, "%0j" prints the day of year as
 *	001-365.
 *
 *	%+ and %- direct Pcal to substitute the following/previous month/year
 *	in the following [bBmyY] specifier; for example, %+B prints the name
 *	of the next month.
 *
 *	Simple cpp-like functionality is provided.  The date file may include
 *	the following commands, which work like their cpp counterparts:
 *
 *		define <sym>
 *		undef <sym>
 *
 *		if{n}def <expr>
 *		   ...
 *		{ else
 *		   ... }
 *		endif
 *
 *		include <file>
 *
 *	Note that these do not start with '#', which is reserved as a comment
 *	character.
 *
 *	<sym> is a symbol name consisting of a letter followed by zero or
 *	more letters, digits, or underscores ('_').  Symbol names are always
 *	treated in a case-insensitive manner.
 *
 *	<expr> is an expression consisting of symbol names joined by the logical
 *	operators (in order of precedence, high to low) '!' (unary negate), '&'
 *	(and), '^' (exclusive or), and '|' (inclusive or).  '&&' and '||' are
 *	accepted as synonyms for '&' and '|' respectively; the order of
 *	evaluation may be altered by the use of parentheses.  A symbol whose
 *	name is currently defined evaluates to TRUE; one whose name is not
 *	currently defined evaluates to FALSE.  Thus "ifdef A | B | C" is TRUE
 *	if any of the symbols A, B, and C is currently defined, and
 *	"ifdef A & B & C" is TRUE if all of them are.
 *
 *	"ifndef A | B | C" is equivalent to "ifdef !(A | B | C)" (or, using
 *	DeMorgan's Law, "ifdef !A & !B & !C") - in other words, TRUE if none of
 *	the symbols A, B, and C is currently defined.
 *
 *	"define" alone deletes all the current definitions; "ifdef" alone is
 *	always false; "ifndef" alone is always true.
 *
 *	The file name in the "include" directive may optionally be surrounded
 *	by "" or <>.  In any case, path names are taken to be relative to
 *	the location of the file containing the "include" directive.
 *
 *	
 *	Moon file syntax:
 *	
 *	The user may enter the dates and (optionally) times of quarter
 *	moons (from a reliable source such as an almanac or astronomical
 *	table) into a file called .moonXX (moonXX.dat on VMS), where XX is
 *	the last two digits of the year.  If such a file exists (in the
 *	same directory as the date file), Pcal will interpolate the phase
 *	of the moon from the information in this file instead of using the
 *	default algorithm.
 *	
 *	(Pcal originally used an extremely simplistic moon phase algorithm;
 *	the moon file was added to v4.0 to enable Pcal to interpolate the
 *	phase of the moon from the [presumably more accurate] information
 *	within.  More recently, the original moon phase algorithm was
 *	superseded by an astronomer-strength version, largely obviating
 *	the need for a moon file; however, it will continue to be
 *	supported for the foreseeable future.)
 *
 *	Entries in the moon file must conform to the following syntax:
 *	
 *	  if -A flag (American date formats) specified:
 *	    <quarter> <month><sep><day> {<hour><sep><min>}
 *	
 *	  if -E flag (European date formats) specified:
 *	    <quarter> <day><sep><month> {<hour><sep><min>}
 *	
 *	where
 *	
 *	  <quarter> := "nm", "fq" or "1q", "fm", "lq" or "3q" (new
 *	  	     moon, first quarter, full moon, last quarter)
 *	  <hour>    := number 0-23 (24-hour clock)
 *	  <min>     := number 0-59
 *	
 *	This file must contain entries for all quarter moons in the year,
 *	in chronological order; if any errors are encountered, Pcal will
 *	revert to using its default algorithm.
 *	
 *	As in the date file, comments start with '#' and run through
 *	end-of-line.  
 */


/*
 * Standard headers:
 */

#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <string.h>

/*
 * Pcal-specific definitions:
 */

#define MAIN_MODULE	1
#include "pcaldefs.h"
#include "pcalglob.h"
#include "pcallang.h"

/*
 * Globals:
 */

static int init_month, init_year, nmonths;


/*
 * Main program - parse and validate command-line arguments, open files,
 * generate PostScript boilerplate and code to generate calendars.
 *
 * Program structure:
 *
 * main() looks for the environment variable (global symbol on VMS) PCAL_OPTS
 * and, if present, calls get_args() to parse it.  It then calls get_args()
 * again to parse the command line for the date file name, any options to be
 * in effect prior to reading the date file, and any numeric arguments (month,
 * year, number of months).  It then calls read_datefile() to read and parse
 * the date file; any "opt" lines present will override the defaults for the
 * command-line flags.  It then calls get_args() yet again to process the
 * remaining command-line flags, which in turn override any specified earlier.
 *
 * main() then attempts to open the output file (if any), and, if successful,
 * calls write_psfile() to generate the PostScript output.
 *
 */
#ifdef PROTOS
int main(int argc,
	 char **argv)
#else
int main(argc, argv)
	int argc;
	char **argv;
#endif
{
	FILE *dfp = NULL;		/* date file pointer */
	char *p, *pathlist[10];
	char tmp[STRSIZ], progpath[STRSIZ];

	INIT_COLORS;		/* copy default_color to day_color */

	/* extract root program name and program path - note that some
	 * systems supply the full pathname and others just the root
	 */

	strcpy(progname, **argv ? *argv : "pcal");

	if ((p = strrchr(progname, END_PATH)) != NULL)
		strcpy(progname, ++p);
	if ((p = strchr(progname, '.')) != NULL)
		*p = '\0';

	mk_path(progpath, find_executable(*argv));

	/* get version from VERSION_STRING (for use in PostScript comment) */
	strcpy(tmp, VERSION_STRING + 4);
	p = strchr(tmp, ' ') + 1;	/* skip program name */
	*strchr(p, ' ') = '\0';		/* terminate after version */
	strcpy(version, p);

	/*
	 * Get the arguments from a) the environment variable, b) "opt" lines
	 * in the date file, and c) the command line, in that order
	 */

	/* parse environment variable PCAL_OPTS as a command line */

	if ((p = getenv(PCAL_OPTS)) != NULL) {
		strcpy(lbuf, "pcal ");		/* dummy program name */
		strcat(lbuf, p);
		(void) loadwords();		/* split string into words */
		if (! get_args(words, P_ENV, PCAL_OPTS)) {
			usage(stderr, FALSE);
			exit(EXIT_FAILURE);
		}
	}

	/* parse command-line arguments once to find name of date file, etc. */

	if (!get_args(argv, P_CMD1, NULL)) {
		usage(stderr, FALSE);
		exit(EXIT_FAILURE);
	}

	/* Attempt to open the date file as specified by the [-e | -f] flags */

	switch (datefile_type) {
	case NO_DATEFILE:
		dfp = NULL;
		break;

	case USER_DATEFILE:	
		/* Attempt to open user-specified calendar file: search
		 * first in PCAL_DIR (current directory if not defined)
		 * and then in the directory where the Pcal executable
		 * lives.  It is a fatal error if the user-specified
		 * date file cannot be found.
		 */
		pathlist[0] = (p = trnlog(PCAL_DIR)) ? p : "";
		pathlist[1] = progpath;
		pathlist[2] = NULL;
		
		strcpy(tmp, datefile);	/* save original name for error msg */

		if ((dfp = alt_fopen(datefile, tmp, pathlist, "r")) == NULL) {
			FPR(stderr, E_FOPEN_ERR, progname, tmp);
			exit(EXIT_FAILURE);
		}
		break;

	case SYS_DATEFILE:
		/* Attempt to open system-specified calendar file: search
		 * first in PCAL_DIR or HOME_DIR (current directory if
		 * neither is defined) and then in the directory where
		 * the Pcal executable lives.  It is not an error if the
		 * system-specified date file cannot be found; Pcal will
		 * simply generate an empty calendar.
		 */
		pathlist[0] = ((p = trnlog(PCAL_DIR)) ||
			       (p = trnlog(HOME_DIR))) ? p : "";
		pathlist[1] = progpath;
		pathlist[2] = NULL;
		
		dfp = alt_fopen(datefile, DATEFILE, pathlist, "r");

		/* if the date file has not been found and ALT_DATEFILE is
		 * defined, search same paths for ALT_DATEFILE before
		 * giving up
		 */
#ifdef ALT_DATEFILE
		if (!dfp)
			dfp = alt_fopen(datefile, ALT_DATEFILE, pathlist, "r");
#endif
		break;
	}

	/* read the date file (if any) and build internal data structure */

	if (dfp) {
		curr_year = init_year;
		read_datefile(dfp, datefile);
		fclose(dfp);
	} else
		datefile[0] = '\0';		/* for PostScript comment */

	/* reparse command line - flags there supersede those in date file */

	(void) get_args(argv, P_CMD2, NULL);

	/* done with the arguments and flags - try to open the output file */

	if (*outfile && freopen(outfile, "w", stdout) == (FILE *) NULL) {
		FPR(stderr, E_FOPEN_ERR, progname, outfile);
		exit(EXIT_FAILURE);
	}

	/* generate the PostScript code (cf. writefil.c) */

	write_psfile(init_month, init_year, nmonths);
	
	cleanup();		/* free allocated data */

	/* if output was written to a non-obvious location, tell user where */

#ifdef DEFAULT_OUTFILE
	FPR(stderr, I_OUT_NAME, progname, outfile);
#endif

	exit(EXIT_SUCCESS);
}


/*
 * set_color - set one or all weekdays to print in black or gray
 */
#ifdef PROTOS
void set_color(char *day,
	       int col)
#else
void set_color(day, col)
	char *day;		/* weekday name (or "all") */
	int  col;		/* select black or gray */
#endif
{
	int i;

	if (ci_strncmp(day, ALL, strlen(ALL)) == 0)	/* set all days */
		for (i = 0; i < 7; i++)
			day_color[i] = col;
	else						/* set single day */
		if ((i = get_weekday(day, FALSE)) != NOT_WEEKDAY)
			day_color[i] = col;

}


/*
 * get_flag() - look up flag in flag_tbl; return pointer to its entry
 * (NULL if not found)
 */
#ifdef PROTOS
FLAG_USAGE *get_flag(char flag)
#else
FLAG_USAGE *get_flag(flag)
	char flag;
#endif
{
	FLAG_USAGE *pflag;

	for (pflag = flag_tbl; pflag->flag; pflag++)
		if (flag == pflag->flag)
			return pflag;

	return flag ? NULL : pflag;		/* '\0' is a valid flag */
}


/*
 * get_args - walk the argument list, parsing all arguments but processing only
 * those specified (in flag_tbl[]) to be processed this pass.
 */
#ifdef PROTOS
int get_args(char **argv,
	     int curr_pass,
	     char *where)
#else
int get_args(argv, curr_pass, where)
	char **argv;		/* argument list */
	int  curr_pass;		/* current pass (P_xxx) */
	char *where;		/* for error messages */
#endif
{
	char *parg, *opt, *p;
	FLAG_USAGE *pflag, *pf;
	int i, flag;
	long tmp;			/* for getting current month/year */
	struct tm *p_tm;
	int badopt = FALSE;		/* flag set if bad param   */
	int nargs = 0;			/* count of non-flag args  */
	int numargs[MAXARGS];		/* non-flag (numeric) args */

/*
 * If argument follows flag (immediately or as next parameter), return
 * pointer to it (and bump argv if necessary); else return NULL
 */
#define GETARG() (*(*argv + 2) ? *argv + 2 : \
		  (*(argv+1) && **(argv+1) != '-' ? *++argv : NULL))

/*
 * Must parse numeric parameters on both command-line passes: before reading
 * datefile (in order to set default year) and again after reading datefile
 * (in order to set the default first month to January if -w flag was set
 * inside the datefile)
 */
#define PARSE_NUM_PARAMS	(curr_pass == P_CMD1 || curr_pass == P_CMD2)

	/* Walk argument list, ignoring first element (program name) */

 	while (opt = *++argv) {

		/* Assume that any non-flag argument is a numeric argument */
		if (*opt != '-') {
		    	if (PARSE_NUM_PARAMS && nargs < MAXARGS) {
				if (! IS_NUMERIC(opt))
					goto bad_par;
				numargs[nargs++] = atoi(opt);
			}
			continue;
		}

		/* Check that flag is a) legal, and b) to be processed this pass */

		if (! (pflag = get_flag(flag = opt[1])) )
			goto bad_par;

		/* get optional argument even if flag not processed this pass */

		parg = pflag->has_arg ? GETARG() : NULL;

		if (! (pflag->passes & curr_pass)) {	/* skip flag this pass? */
			if (curr_pass == P_OPT)
				FPR(stderr, E_FLAG_IGNORED, progname, flag,
					DATE_FILE, where);
			continue;
		}

		switch (flag) {

		case F_INITIALIZE:	/* reset all flags to defaults */

			/* set up a command line to reset all of the
			 * flags; call get_args() recursively to parse it
			 * (note that some of the flags must be reset
			 * explicitly, as no command-line flags exist to
			 * reset them)
			 */

			/* reset flags described above */
			julian_dates  = JULIAN_DATES;
			draw_moons    = DRAW_MOONS;
			do_whole_year = FALSE;
			blank_boxes   = FALSE;
			num_style     = NUM_STYLE;

			/* select program default for landscape/portrait
			 * mode and US/European date styles
			 */
			sprintf(lbuf, "pcal -%c -%c",
#if (ROTATE == LANDSCAPE)
				F_LANDSCAPE,
#else
				F_PORTRAIT,
#endif
#if (DATE_STYLE == USA_DATES)
				F_USA_DATES);
#else
				F_EUR_DATES);
#endif
			p = lbuf + strlen(lbuf);

			/* all other flags take arguments and are reset
			 * by specifying the flag without an argument
			 */
			for (pf = flag_tbl; pf->flag; pf++)
				if ((pf->passes & curr_pass) && pf->has_arg) {
					sprintf(p, " -%c", pf->flag);
					p += strlen(p);
				}

			/* split new command line into words; parse it */
			(void) loadwords();
			(void) get_args(words, curr_pass, NULL);

			break;

		case F_BLACK_DAY:	/* print day in black or gray */
		case F_GRAY_DAY:
			if (parg)
				set_color(parg, flag == F_BLACK_DAY ?
						  BLACK : GRAY);
			else
				INIT_COLORS;	/* reset to defaults */
			break;

 		case F_DAY_FONT:	/* specify alternate day font */
			strcpy(dayfont, parg ? parg : DAYFONT);
 			break;

		case F_NOTES_FONT:	/* specify alternate notes font */
			strcpy(notesfont, parg ? parg : NOTESFONT);
			break;

 		case F_TITLE_FONT:	/* specify alternate title font */
			strcpy(titlefont, parg ? parg : TITLEFONT);
 			break;

		case F_EMPTY_CAL:	/* generate empty calendar */
			datefile_type = NO_DATEFILE;
			strcpy(datefile, "");
			break;

		case F_DATE_FILE:	/* specify alternate date file */
			datefile_type = parg ? USER_DATEFILE : SYS_DATEFILE;
			strcpy(datefile, parg ? parg : "");
			break;

		case F_OUT_FILE:	/* specify alternate output file */
			strcpy(outfile, parg ? parg : OUTFILE);
			break;

		case F_LANDSCAPE:	/* generate landscape calendar */
 			rotate = LANDSCAPE;
			strcpy(xsval, XSVAL_L);
			strcpy(ysval, YSVAL_L);
			strcpy(xtval, XTVAL_L);
			strcpy(ytval, YTVAL_L);
 			break;
 
		case F_PORTRAIT:	/* generate portrait calendar */
 			rotate = PORTRAIT;
			strcpy(xsval, XSVAL_P);
			strcpy(ysval, YSVAL_P);
			strcpy(xtval, XTVAL_P);
			strcpy(ytval, YTVAL_P);
 			break;

		case F_HELP:		/* request "help" message or version */
		case F_VERSION:
			FPR(stdout, "%s\n", VERSION_STRING + 4);
			if (flag == F_HELP)
				usage(stdout, TRUE);
			exit(EXIT_SUCCESS);
			break;

		case F_MOON_4:		/* draw four moons */
		case F_MOON_ALL:	/* draw a moon for each day */
			draw_moons = flag == F_MOON_ALL ? ALL_MOONS : SOME_MOONS;
			break;

		case F_DEFINE:		/* define preprocessor symbol */
			(void) do_define(parg);
			break;

		case F_UNDEF:		/* undef preprocessor symbol */
			(void) do_undef(parg);
			break;

		case F_L_FOOT:		/* specify alternate left foot */
			strcpy(lfoot, parg ? parg : LFOOT);
			break;

		case F_C_FOOT:		/* specify alternate center foot */
			strcpy(cfoot, parg ? parg : CFOOT);
			break;

 		case F_R_FOOT:		/* specify alternate right foot */
			strcpy(rfoot, parg ? parg : RFOOT);
			break;

		case F_FIRST_DAY:	/* select starting day of week */
			if (parg) {
				if ((i = get_weekday(parg, FALSE)) != NOT_WEEKDAY)
					first_day_of_week = i;
			}
			else
				first_day_of_week = FIRST_DAY;
			break;

		case F_USA_DATES:	/* select American date style */
		case F_EUR_DATES:	/* select European date style */
			date_style = flag == F_USA_DATES ? USA_DATES : EUR_DATES;
			break;

		case F_X_TRANS:		/* set x-axis translation factor */
			strcpy(xtval, parg ? parg :
				(rotate == LANDSCAPE ? XTVAL_L : XTVAL_P));
			break;

		case F_Y_TRANS:		/* set y-axis translation factor */
			strcpy(ytval, parg ? parg :
				(rotate == LANDSCAPE ? YTVAL_L : YTVAL_P));
			break;

		case F_X_SCALE:		/* set x-axis scaling factor */
			strcpy(xsval, parg ? parg :
				(rotate == LANDSCAPE ? XSVAL_L : XSVAL_P));
			break;

		case F_Y_SCALE:		/* set y-axis scaling factor */
			strcpy(ysval, parg ? parg :
				(rotate == LANDSCAPE ? YSVAL_L : YSVAL_P));
			break;

		case F_JULIAN:
		case F_JULIAN_ALL:
			julian_dates = flag == F_JULIAN_ALL ? ALL_JULIANS :
							      SOME_JULIANS;
			break;

		case F_WHOLE_YEAR:
			do_whole_year = TRUE;
			break;

		case F_BLANK_BOXES:
			blank_boxes = TRUE;
			break;

		case F_OUTLINE:
		case F_OUTLINE_GRAY:
			num_style = flag == F_OUTLINE ? OUTLINE_NUMS :
							FILLED_NUMS;
			break;

		case '-' :		/* accept - and -- as dummy flags */
		case '\0':
			break;

		default:		/* missing case label if reached!!! */

bad_par:				/* unrecognized parameter */

			FPR(stderr, E_ILL_OPT, progname, opt);
			if (where)
				FPR(stderr, E_ILL_OPT2,
					curr_pass == P_ENV ? ENV_VAR :
					curr_pass == P_OPT ? DATE_FILE : "",
					where);
			FPR(stderr, "\n");
			badopt = TRUE;
			break;
		}
        }

	if (! PARSE_NUM_PARAMS)
		return !badopt;		/* return TRUE if OK, FALSE if error */

	/* Validate non-flag (numeric) parameters */

	switch (nargs) {
	case 0:		/* no arguments - print current month and/or year */
		time(&tmp);
		p_tm = localtime(&tmp);
		init_month = do_whole_year ? JAN : p_tm->tm_mon + 1;
		init_year = p_tm->tm_year;
		nmonths = 1;
		break;			
	case 1:		/* one argument - print entire year */
		init_month = JAN;
		init_year = numargs[0];
		nmonths = 12;
		break;
	default:	/* two or three arguments - print one or more months */
		init_month = numargs[0];
		init_year = numargs[1];
		nmonths = nargs > 2 ? numargs[2] : 1;
		break;
	}

	if (nmonths < 1)		/* ensure at least one month */
		nmonths = 1;

	/* check range of month and year */

	if (init_month < JAN || init_month > DEC) {
		FPR(stderr, E_ILL_MONTH, progname, init_month, JAN, DEC);
		badopt = TRUE;
	}
	
	if (init_year > 0 && init_year < 100)	/* treat nn as 19nn */
		init_year += CENTURY;
	
	if (init_year < MIN_YR || init_year > MAX_YR) {
		FPR(stderr, E_ILL_YEAR, progname, init_year, MIN_YR, MAX_YR);
		badopt = TRUE;
	}

	return !badopt;		/* return TRUE if OK, FALSE if error */
}



/*
 * color_msg - return character string explaining default day colors
 */
#ifdef PROTOS
char *color_msg(void)
#else
char *color_msg()
#endif
{
	int i, ngray = 0, others;
	static char msg[80];

	for (i = SUN; i <= SAT; i++)	/* count gray weekdays */
		if (default_color[i] == GRAY)
			ngray++;

	if (ngray == 0 || ngray == 7) {		/* all same color? */
		sprintf(msg, COLOR_MSG_1, ngray ? W_GRAY : W_BLACK);
		return msg;
	}

	others = ngray <= 3 ? BLACK : GRAY;	/* no - get predominant color */
	msg[0] = '\0';
	for (i = SUN; i <= SAT; i++)
		if (default_color[i] != others) {
			strncat(msg, days[i], MIN_DAY_LEN);
			strcat(msg, "/");
		}
	LASTCHAR(msg) = ' ';

	sprintf(msg + strlen(msg), COLOR_MSG_2,
		others == BLACK ? W_GRAY : W_BLACK,
                others == BLACK ? W_BLACK : W_GRAY);
	return msg;
}


/*
 * usage - print message explaining correct usage of the command-line
 * arguments and flags.  If "fullmsg" is true, print associated text
 */
#ifdef PROTOS
void usage(FILE *fp,
	   int fullmsg)
#else
void usage(fp, fullmsg)
	FILE *fp;	/* destination of usage message */
	int fullmsg;	/* print complete message? */
#endif
{
	FLAG_MSG *pflag;
	PARAM_MSG *ppar;
	DATE_MSG *pdate;
	char buf[30], *p, flag, *meta;
	int nchars, first, i, indent, n;

	sprintf(buf, "%s: %s", W_USAGE, progname);
	nchars = indent = strlen(buf);
	first = TRUE;
	meta = p = NULL;
	FPR(fp, "\n%s", buf);

	/* loop to print command-line syntax message (by group of flags) */

	for (pflag = flag_msg; (flag = pflag->flag) != '\0'; pflag++) {
		if (flag == '\n') {		/* end of group? */
			if (p)
				*p = '\0';
			if (meta) {		/* append metavariable name */
				strcat(buf, " ");
				strcat(buf, meta);
			}
			strcat(buf, "]");
			n = strlen(buf);
			if (nchars + n > SCREENWIDTH) {	/* does it fit on line? */
				FPR(fp, "\n");		/* no - start new one */
				for (nchars = 0; nchars < indent; nchars++)
					FPR(fp, " ");
			}
			FPR(fp, "%s", buf);
			nchars += n;
			first = TRUE;
		}
		else if (flag != ' ') {		/* accumulate flags for group */
			if (first) {
				sprintf(buf, " [");
				p = buf + strlen(buf);
			}
			else
				*p++ = '|';
			*p++ = '-';
			*p++ = flag;
			meta = pflag->meta;	/* save metavariable name */
			first = FALSE;
		}
	}

	/* loop to print selected numeric parameter descriptions */

	for (i = 0; i < PARAM_MSGS; i++) {
		sprintf(buf, " [%s]%s", param_msg[i].desc,
			i < PARAM_MSGS - 1 ? " |" : "");
		n = strlen(buf);
		if (nchars + n > SCREENWIDTH) {	/* does it fit on line? */
			FPR(fp, "\n");		/* no - start new one */
			for (nchars = 0; nchars < indent; nchars++)
				FPR(fp, " ");
		}
		FPR(fp, "%s", buf);
		nchars += n;
	}

	FPR(fp, "\n\n");
	if (! fullmsg) {
		FPR(fp, USAGE_MSG, progname, F_HELP);
		return;
	}
	
	/* loop to print the full flag descriptions */

	for (pflag = flag_msg; (flag = pflag->flag) != '\0'; pflag++) {
		if (flag == '\n') {	/* newline?  print and quit */
			FPR(fp, "\n");
			continue;
		}
		p = buf;		/* copy flag and metavariable to buffer */
		if (flag != ' ')
			*p++ = '-';
	/* special hack for VMS - surround upper-case flags in quotes */
#ifdef VMS
		if (isupper(flag)) {
			*p++ = '"';
			*p++ = flag;
			*p++ = '"';
		}
		else
			*p++ = flag;
#else
		*p++ = flag;
#endif
		*p = '\0';
		if (pflag->meta)
			sprintf(p, " <%s>", pflag->meta);
		FPR(fp, "\t%-16.16s", buf);
		if (pflag->text)
			FPR(fp, "%s", pflag->text);

		/* print default value if specified */
		if (pflag->def)
			FPR(fp, " (%s: %s)", W_DEFAULT, pflag->def[0] ? pflag->def : "\"\"" );
		FPR(fp, "\n");

		/* special case - print color messages after F_GRAY_DAY */
		if (flag == F_GRAY_DAY)
			FPR(fp, "\t\t\t  (%s: %s)\n", W_DEFAULT, color_msg());

	}
	
	/* now print the information about the numeric parameters */

	for (ppar = param_msg; ppar->desc; ppar++)
		FPR(fp, "\t%-16.16s%s\n", ppar->desc, ppar->text);
	
	/* print the date file syntax message */

	FPR(fp, "\n");
	for (pdate = date_msg; *pdate; pdate++)
		FPR(fp, "\t%s\n", *pdate);

}


/*
 * alt_fopen - attempt to open a file in one of several paths in a
 * NULL-terminated path list.  If successful, return (opened) file pointer
 * and fill in full path name; if not, return NULL
 */
#ifdef PROTOS
FILE *alt_fopen(char *fullpath, char *name, char *pathlist[], char *access) 
#else
FILE *alt_fopen(fullpath, name, pathlist, access)
	char *fullpath;		/* full path name (output) */
	char *name;		/* base name (or full path spec) */
	char *pathlist[];	/* NULL-terminated path list */
	char *access;		/* permission requested */
#endif
{
	char **path;
	FILE *fp;

	for (path = pathlist; *path; path++) {
		mk_filespec(fullpath, *path, name);
		if ((fp = fopen(fullpath, access)) != NULL)
			return fp;
		}

	fullpath[0] = '\0';		/* file not found */
	return NULL;
}

