/*
 * writefil.c - Pcal routines concerned with writing the PostScript output
 *
 * Contents:
 *
 *		def_footstring
 *		expand_fmt
 *		find_daytext
 *		find_holidays
 *		print_julian_info
 *		print_month
 *		print_moon_info
 *		print_text
 *		print_word
 *		write_psfile
 *
 * Revision history:
 *
 *	4.1	AWR	08/16/91	Support -G flag (outlined gray dates)
 *
 *	4.02	AWR	07/02/91	Added "%" expansions in text strings
 *					(cf. expand_fmt())
 *
 *	4.0	AWR	01/28/91	Support -B, -w flags and moon file
 *
 *			01/15/91	Extracted from pcal.c
 *
 */

/*
 * Standard headers:
 */

#include <stdio.h>
#include <ctype.h>

/*
 * Pcal-specific definitions:
 */

#include "pcaldefs.h"
#include "pcalglob.h"
#define  WRITEFIL		/* to get ordinal_suffix() from pcallang.h */
#include "pcallang.h"
#include "pcalinit.h"		/* PostScript boilerplate */

/*
 * Macros:
 */

/* convert "ph" (approximate quarter moon) to exact quarter moon */
#define EXACT_QUARTER(ph)	((((int) (((ph) + .05) * 4.0)) % 4) / 4.0)

/* make sure printf() doesn't round "ph" up to 1.0 when printing it */
#define PRT_TWEAK(ph)		((ph) >= 0.9995 ? 0.0 : (ph))

/*
 * Globals:
 */

/* order of following strings must conform to #define's in pcaldefs.h (q.v.) */
static char *cond[3] = {"false", "true", "(some)"};
static char *gray[3] = {"(gray)", "(outline)", "(outline_gray)"};

static int this_day, this_month, this_year;


/*
 * write_psfile - write PostScript code to stdout
 *
 * The actual output of the PostScript code is straightforward.  This routine
 * writes a PostScript header followed by declarations of all the PostScript
 * variables affected by command-line flags and/or language dependencies.  It
 * the generates the PostScript boilerplate generated from pcalinit.ps, and
 * finally calls print_month() to generate the PostScript code for each
 * requested month.
 */
#ifdef PROTOS
void write_psfile(int month,
		  int year,
		  int nmonths)
#else
void write_psfile(month, year, nmonths)
	int month;			/* starting month   */
	int year;			/* starting year    */
	int nmonths;			/* number of months */
#endif
{
	int i, ngray;
	char **ap, tmp[20];

	/*
	 * Write out PostScript prolog
	 */

 	PRT("%%!\n%%\t");
	PRT(VERSION_MSG, progname, version);
	if (*datefile)
		PRT(DATEFILE_MSG, datefile);
	PRT("\n%%\n");

	/* font names */

	PRT("/titlefont /%s def\n/dayfont /%s def\n/notesfont /%s def\n",
	    titlefont, dayfont, notesfont);

	/* month names */

	PRT("/month_names [");
	for (i = JAN; i <= DEC; i++) {
		PRT(i % 6 == 1 ? "\n\t" : " ");
		(void) print_word(months[i-1]);
	}
	PRT(" ] def\n");

	/* day names - abbreviate if printing entire year on page */

	PRT("/day_names [");
	for (i = SUN; i <= SAT; i++) {
		PRT(i % 6 == 0 && ! do_whole_year ? "\n\t" : " ");
		strcpy(tmp, days[(i + first_day_of_week) % 7]);
		if (do_whole_year)
			tmp[MIN_DAY_LEN] = '\0';
		(void) print_word(tmp);
		}
	PRT(" ] def\n");

	/* title for "notes" box */

	PRT("/notesheading (");
	PUTSTR(NOTES_HDR);
	PRT(") def\n");

	/* line separator */

	PRT("/linesep ");
	print_word(LINE_SEP);
	PRT(" def\n");

	/* colors (black/gray) to print weekdays and holidays */

	PRT("/day_gray [");
	for (ngray = 0, i = SUN; i <= SAT; ngray += day_color[i++] == GRAY)
		PRT(" %s", cond[day_color[(i + first_day_of_week) % 7]]);
	PRT(" ] def\n");
	PRT("/holiday_gray %s def\n", cond[ngray <= 3]);

 	/* rotation, scale, and translate values. */
 
 	PRT("/rval %d def\n", rotate);
 	PRT("/xsval %s def\n/ysval %s def\n", xsval, ysval);
 	PRT("/xtval %s def\n/ytval %s def\n", xtval, ytval);

	/* moon, Julian date, box fill, and outline flags */

	PRT("/draw-moons %s def\n", cond[draw_moons]);
	PRT("/julian-dates %s def\n", cond[julian_dates]);
	PRT("/fill-boxes %s def\n", cond[! blank_boxes]);
	PRT("/logical_gray %s def\n", gray[num_style]);

	/* PostScript boilerplate (part 1 of 1) */

	for (ap = header; *ap; ap++)
		PRT("%s\n", *ap);
	PRT("\n");

	/*
	 * Write out PostScript code to print calendars
	 */

	if (do_whole_year)	/* round up to multiple of 12 months */
		nmonths = ((nmonths + 11) / 12) * 12;

	for (this_month = month, this_year = year; nmonths--; ) {
		print_month(this_month, this_year);
		this_year = NEXT_YEAR(this_month, this_year);
		this_month = NEXT_MONTH(this_month, this_year);
	}

}


/*
 * low-level utilities for PostScript generation
 */

/*
 * expand_fmt - expand a strftime-like date format string; pcal supports
 * %[aAbBdjmUWyY] from strftime() plus %l and prefixes [0o+-] (see below);
 * returns pointer to character following end of string
 */
#ifdef PROTOS
char *expand_fmt(char *p)
#else
char *expand_fmt(p)
	char *p;		/* character following percent sign */
#endif
{
	char c, tmp[20];
	static char *prefixes = "0o+-";
	int firstday, wkday;
	int adjust = 0, strip_lz = TRUE, ordinal = FALSE, prev_num = 0;

	do {
		switch (c = *p++) {
		case 'a':	/* %a : abbreviated weekday */
		case 'A':	/* %A : full weekday */
			wkday = calc_weekday(this_month, this_day, this_year);
			strcpy(tmp, days[wkday]);
			if (c == 'a')
				tmp[MIN_DAY_LEN] = '\0';
			printf("%s", tmp);
			break;

		case 'b':	/* %b : abbreviated month name */
		case 'B':	/* %B : full month name */
			strcpy(tmp, months[(this_month + adjust + 11) % 12]);
			if (c == 'b')
				tmp[MIN_MONTH_LEN] = '\0';
			printf("%s", tmp);
			break;

		case 'd':	/* %d : day of month (01-31) */
			printf(strip_lz ? "%d" : "%02d", prev_num = this_day);
			break;

		case 'j':	/* %j : day of year (001-366) */
			prev_num = DAY_OF_YEAR(this_month, this_day, this_year);
			printf(strip_lz ? "%d" : "%03d", prev_num);
			break;

		case 'l':	/* %l : days left in year (000-365) (NEW) */
			prev_num = YEAR_LEN(this_year) - DAY_OF_YEAR(this_month,					this_day, this_year);
			printf(strip_lz ? "%d" : "%03d", prev_num);
			break;

		case 'm':	/* %m : month (01-12) */
			prev_num = (this_month + adjust + 11) % 12 + 1;
			printf(strip_lz ? "%d" : "%02d", prev_num);
			break;

		/* %U considers the first "logical Sunday" (the first day
		 * of the week as printed - cf. the -F option) of the year
		 * to be the start of week 1; %W uses the first "logical
		 * Monday" instead.
		 */
		case 'U':	/* %U : week number (00-53) */
		case 'W':	/* %W : week number (00-53) */
			firstday = ((c == 'W' ? 15 : 14) -
					START_DAY(JAN, this_year)) % 7 + 1;
			prev_num = (DAY_OF_YEAR(this_month, this_day,
					this_year) - firstday + 7) / 7;
			printf(strip_lz ? "%d" : "%02d", prev_num);
			break;

		case 'y':	/* %y : year w/o century (00-99) */
			printf("%02d", prev_num = (this_year + adjust) % 100);
			break;

		case 'Y':	/* %Y : year w/century */
			printf("%d", prev_num = this_year + adjust);
			break;

		/* prefix flags [o0+-] : set flags for next pass */

		case 'o':	/* %o : ordinal suffix (NEW) */
			ordinal = TRUE;
			break;

		case '0':	/* %0 : add leading zeroes (NEW) */
			strip_lz = FALSE;
			break;

		case '+':	/* %+ : increment next month/year (NEW) */
		case '-':	/* %- : decrement next month/year (NEW) */
			adjust = c == '-' ? -1 : 1;
			break;

		case '\0':	/* accidental end-of-string */
		case ' ':
			return p - 1;

		default:	/* other - just print it */
			PUTCHAR(c);
			break;
		};

	} while (strchr(prefixes, c) != NULL);

	if (ordinal && prev_num)
		printf("%s", ordinal_suffix(prev_num));
	return p;

}


/*
 * print_word - print a single word, representing parentheses and non-ASCII
 * characters as octal literals and expanding format specifiers; return pointer
 * to character following word (NULL if no word follows)
 */
#ifdef PROTOS
char *print_word(char *p)
#else
char *print_word(p)
	char *p;
#endif
{
	char c;

	if (*p == '\0' || *(p += strspn(p, WHITESPACE)) == '\0')
		return NULL;

	PRT("(");
	while ((c = *p) && !isspace(c)) {
		if (c == '%' && p[1] != '\0')
			p = expand_fmt(p + 1);
		else {
			PUTCHAR(c);
			p++;
		}
	}
	PRT(")");

	return p;
}


/*
 * print_text - print tokens in text (assumed separated by single blank)
 * in PostScript format
 */
#ifdef PROTOS
void print_text(char *p)
#else
void print_text(p)
	char *p;
#endif
{
	while (p = print_word(p))
		PRT(" ");
}


/*
 * def_footstring - print definition for foot string, again converting 
 * all characters other than letters, digits, or space to octal escape
 * and expanding format specifiers
 */
#ifdef PROTOS
void def_footstring(char *p,
		    char lcr)
#else
void def_footstring(p, lcr)
	char *p;			/* definition */
	char lcr;			/* L, C, or R */
#endif
{
	char c;

	this_day = 1;			/* set default day in foot string */
	PRT("/%cfootstring (", lcr);
	while (c = *p)
		if (c == '%' && p[1] != '\0')
			p = expand_fmt(p + 1);
		else
			PUTCHAR(c), p++;
	PRT(") def\n");
}


/*
 * Routines to extract and print data
 */


/*
 * find_daytext - find and print the day (including notes) or holiday text
 * for the specified month/year
 */
#ifdef PROTOS
void find_daytext(int month,
		  int year,
		  int is_holiday)
#else
void find_daytext(month, year, is_holiday)
	int month, year;
	int is_holiday;		/* TRUE: print holiday text */
#endif
{
	register int day;
	year_info *py;
	month_info *pm;
	register day_info *pd;
	int first;
	char *fcn = is_holiday ? "holidaytext" : "daytext";

	/* if no text for this year and month, return */

	if ((py = find_year(year, FALSE)) == NULL ||
	    (pm = py->month[month-1]) == NULL)
		return;

	/* walk array of day text pointers and linked lists of text */

	for (day = 1; day <= NOTE_DAY; day++) {
		for (pd = pm->day[day-1], first = TRUE;
		     pd;
		     pd = pd->next) {
			if (pd->is_holiday != is_holiday)
				continue;
			if (first) {
				if (day != NOTE_DAY)	/* set up call */
					PRT("%d ", day);
				printf("[ \n");
			}
			else {
				PRT("\n");
				print_word(LINE_SEP);	/* separate lines */
				PRT("\n");
			}
			this_day = day == NOTE_DAY ? 1 : day;
			print_text(pd->text);
			first = FALSE;
		}
		if (! first)		/* wrap up call (if one made) */
			PRT("\n] %s\n", day == NOTE_DAY ? "notetext" : fcn);
	}
}


/*
 * find_holidays - find and print the note block flag and holidays for
 * specified month/year
 */
#ifdef PROTOS
void find_holidays(int month,
		   int year)
#else
void find_holidays(month, year)
	int month, year;
#endif
{
	register int day;
	register unsigned long holidays;
	year_info *py;
	month_info *pm;

	pm = (py = find_year(year, FALSE)) ? py->month[month-1] : NULL;

	if (! do_whole_year)
 		PRT("/note_block %s def\n", cond[pm && pm->day[NOTE_DAY-1]]);

	PRT("/holidays [");	/* start definition of list */

	for (holidays = pm ? pm->holidays : 0, day = 1;
	     holidays;
	     holidays >>= 1, day++)
		if (holidays & 01)
			PRT(" %d", day);

	PRT(" ] def\n");

}


/*
 * print_moon_info - print the information necessary to draw moons.  If
 * printing moons on all days, print the phase for each day; if printing
 * only quarter moons, tweak the phase to an exact quarter (so the icon
 * is printed correctly) and generate a list of the quarter-moon dates
 */
#ifdef PROTOS
void print_moon_info(int month,
		     int year)
#else
void print_moon_info(month, year)
	int month, year;
#endif
{
	int n, ndays, day, is_q;
	char *p;
	unsigned long qdays;
	double phase;

	if (draw_moons == NO_MOONS)
		return;

	/* print the phase of the moon for each day of the month */

	PRT("/moon_phases [\t\t%% from %s\n\t",
		(p = find_moonfile(year)) ? p : "algorithm");

	for (n = 0, qdays = 0L, day = 1, ndays = LENGTH_OF(month, year);
	     day <= ndays;
	     day++) {
		phase = find_phase(month, day, year, &is_q);
		if (draw_moons == SOME_MOONS && is_q)
			phase = EXACT_QUARTER(phase);
		if (draw_moons == ALL_MOONS || is_q)
			PRT("%.3f%s", PRT_TWEAK(phase), ++n % 10 == 0 ?
			    "\n\t" : " ");
		if (is_q)
			qdays |= 1L << (day - 1);
		}
	PRT("] def\n");

	/* if drawing only quarter moons, print days when they occur */

	if (draw_moons == SOME_MOONS) {
		PRT("/quarter_moons [ ");
		for (day = 1; qdays; day++, qdays >>= 1)
			if (qdays & 01)
				PRT("%d ", day);
		PRT("] def\n");
	}
}


/*
 * print_julian_info - print the information necessary to print Julian dates
 */
#ifdef PROTOS
void print_julian_info(int month,
		       int year)
#else
void print_julian_info(month, year)
	int month, year;
#endif
{

	if (julian_dates != NO_JULIANS)
		PRT("/jdstart %d def\n", DAY_OF_YEAR(month, 1, year));
	if (julian_dates == ALL_JULIANS)
		PRT("/yearlen %d def\n", YEAR_LEN(year));
}


/*
 * print_month - generate calendar for specified month/year
 */
#ifdef PROTOS
void print_month(int month,
		 int year)
#else
void print_month(month, year)
	int month, year;
#endif
{
	static int nmonths = 0;

	PRT("/year %d def\n", year);		/* set up year and month */
	PRT("/month %d def\n", month);
	PRT("/startday %d def\n", START_DAY(month, year));
	PRT("/ndays %d def\n", LENGTH_OF(month, year));

	find_holidays(month, year);		/* make list of holidays */

	/* are we printing 12 months per page or only one? */

	if (do_whole_year) {
		/* reset foot strings at start of each page */
		if (nmonths % 12 == 0) {
			def_footstring(lfoot, 'L');
			def_footstring(cfoot, 'C');
			def_footstring(rfoot, 'R');
		}

		PRT("/posn %d def\n", nmonths % 12);	/* location on page */
		PRT("printmonth_%c\n", rotate == LANDSCAPE ? 'l' : 'p');
		if (++nmonths % 12 == 0)
			PRT("showpage\n");
	}
	else {
		/* reset foot strings each month (may change) */
		def_footstring(lfoot, 'L');
		def_footstring(cfoot, 'C');
		def_footstring(rfoot, 'R');

		/* generate lengths of previous and next months */
		PRT("/lndays %d def\n", LENGTH_OF(PREV_MONTH(month, year),
					PREV_YEAR(month, year)));
		PRT("/nndays %d def\n", LENGTH_OF(NEXT_MONTH(month, year),
					NEXT_YEAR(month, year)));

		print_julian_info(month, year);		/* Julian date info */
		print_moon_info(month, year);		/* moon info */

		PRT("printmonth\n");
		find_daytext(month, year, TRUE);	/* holiday text */
		find_daytext(month, year, FALSE);	/* day and note text */
		PRT("showpage\n");
	}
}
